/*	a r r a y c o n v s
	apl array conversion routines.
*/
#include <apl.h>
#include <stdlib.h>
#include <math_iso99.h>
#include <decf.h>
#include <string.h>
#include <support.h>
#define	UNS	unsigned
#define	UCHAR	unsigned char

static	ARRAY	*dest;
static	BOUND	els;

/*	m k _ p t
	make an ARRAY, setting up dest (ptr), els (bnd).
*/
static void *mk_pt(UNS eltype, UNS rank, BOUND *pshape)
{
	els = arraybound(dest = mkarray(eltype, rank, pshape));
	return arraystart(dest);
}

static ARRAY *adomx(ARRAY *src)
	{
	domain();
	return 0; // keep compiler happy
	}

static ARRAY *anonx(ARRAY *src)
	{
	nonce();
	return 0; // keep compiler happy
	}

/*	a g e t _ a
	make copy of array and return pointer to copy.
*/

static ARRAY *aget_a(ARRAY *src)
{
	size_t	bytes;
	
	dest = (ARRAY *)malloc(bytes = src->length * sizeof(MEMORYINT));
	
	if(dest)
		memcpy(dest, src, bytes);

	return(dest);
}

ARRAY	*sget_b(ARRAY *src) // get APLSINT array from APLBOOL array. 
{
	apluns8		*from;
	aplint8		*to;
	BOUND		bnd;
	apluns8		boolmask;

	from = arraystart(src);
	to = mk_pt(APLSINT, src->rank, src->shape);
	bnd = els;

	for(--from, boolmask = 0; bnd--;)
		{
		if(! (boolmask >>= 1))
			{
			boolmask = 0x80;
			++from;
			}
		*to++ = 0 != (boolmask & *from);
		}
	return(dest);
}

ARRAY	*iget_b(ARRAY *src)	// get APLINTG array from APLBOOL array. 
{
	apluns8		*from;
	aplint16	*to;
	BOUND		bnd;
	apluns8		boolmask;

	from = arraystart(src);
	to = mk_pt(APLINTG, src->rank, src->shape);
	bnd = els;
	
	for(--from, boolmask = 0; bnd--;)
		{
		if(! (boolmask >>= 1))
			{
			boolmask = 0x80;
			++from;
			}
		*to++ = 0 != (boolmask & *from);
		}
	return(dest);
}

ARRAY	*lget_b(ARRAY *src) // get APLLONG array from APLBOOL array. 
{
	apluns8		*from;
	aplint32	*to;
	BOUND		bnd;
	apluns8		boolmask;

	from = arraystart(src);
	to = mk_pt(APLLONG, src->rank, src->shape);
	bnd = els;
	
	for(--from, boolmask = 0; bnd--;)
		{
		if(! (boolmask >>= 1))
			{
			boolmask = 0x80;
			++from;
			}
		*to++ = 0 != (boolmask & *from);
		}
	return(dest);
}

ARRAY	*dget_b(ARRAY *src) // get APLDOUB array from APLBOOL array. 
{
	apluns8		*from;
	aplfloat64	*to;
	BOUND		bnd;
	apluns8		boolmask;

	from = (UCHAR *)arraystart(src);
	to = (double *)mk_pt(APLDOUB, src->rank, src->shape);
	bnd = els;

	for(--from, boolmask = 0; bnd--;)
		{
		if(! (boolmask >>= 1))
			{
			boolmask = 0x80;
			++from;
			}
		*to++ = 0 != (boolmask & *from);
		}
	return(dest);
}

static ARRAY *iget_s(ARRAY *src) // get APLINTG array from APLSINT array. 
{
	aplint8		*from;
	aplint16	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLINTG, src->rank, src->shape);
	bnd = els;

	while(bnd--) *to++ = *from++;

	return(dest);
}

static ARRAY *lget_s(ARRAY *src) // get APLLONG array from APLSINT array. 
{
	aplint8		*from;
	aplint32	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLLONG, src->rank, src->shape);
	bnd = els;

	while(bnd--) *to++ = *from++;

	return(dest);
}

static ARRAY *dget_s(ARRAY *src) // get APLDOUB array from APLSINT array. 
{
	aplint8		*from;
	aplfloat64	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDOUB, src->rank, src->shape);
	bnd = els;

	while(bnd--)
		*to++ = *from++;

	return(dest);
}

static ARRAY *lget_i(ARRAY *src) // get APLLONG array from APLINTG array. 
{
	aplint16	*from;
	aplint32	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLLONG, src->rank, src->shape);
	bnd = els;

	while(bnd--) *to++ = *from++;

	return(dest);
}

static ARRAY *dget_i(ARRAY *src) // get APLDOUB array from APLINTG array. 
{
	aplint16	*from;
	aplfloat64	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDOUB, src->rank, src->shape);
	bnd = els;

	while(bnd--)
		*to++ = *from++;

	return(dest);
}

static ARRAY *dget_l(ARRAY *src) // get APLDOUB array from APLLONG array. 
{
	aplint32	*from;
	aplfloat64	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDOUB, src->rank, src->shape);
	bnd = els;

	while(bnd--) 
		*to++ = *from++;

	return(dest);
}

static ARRAY *c8getc16(ARRAY *src) // get APLCHAR16 array from APLCHAR8 array. 
{
	aplWchar8	*from;
	aplWchar16	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLWCHAR16, src->rank, src->shape);
	bnd = els;

	while(bnd--) 
		*to++ = *from++;

	return(dest);
}

static ARRAY *c8getc32(ARRAY *src) // get APLCHAR32 array from APLCHAR8 array. 
{
	aplWchar8	*from;
	aplWchar32	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLWCHAR32, src->rank, src->shape);
	bnd = els;

	while(bnd--) 
		*to++ = *from++;

	return(dest);
}

static ARRAY *c16getc32(ARRAY *src) // get APLCHAR32 array from APLCHAR16 array. 
{
	aplWchar16	*from;
	aplWchar32	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLWCHAR32, src->rank, src->shape);
	bnd = els;

	while(bnd--) 
		*to++ = *from++;

	return(dest);
}

static ARRAY *zget_b(ARRAY *src)	// get APLCMPX from APLBOOL
{
	apluns8		*from;
	aplcmpx128	*to;
	BOUND		bnd;
	apluns8		boolmask;

	from = (apluns8 *)arraystart(src);
	to = (aplcmpx128 *)mk_pt(APLCMPX, src->rank, src->shape);
	bnd = els;

	for(--from, boolmask = 0; bnd--; ++to)
		{
		if(! (boolmask >>= 1))
			{
			boolmask = 0x80;
			++from;
			}
		to->im = 0.0;
		to->re = 0 != (boolmask & *from);
		}
	return(dest);
}

static ARRAY *zget_s(ARRAY *src)	// get APLCMPX from APLSINT
{
	aplint8		*from;
	aplcmpx128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLCMPX, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		to->im = 0.0, to->re = *from;

	return(dest);
}

static ARRAY *zget_i(ARRAY *src)	// get APLCMPX from APLINTG
{
	aplint16	*from;
	aplcmpx128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLCMPX, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		to->im = 0.0, to->re = *from;

	return(dest);
}

static ARRAY *zget_l(ARRAY *src)	// get APLCMPX from APLLONG
{
	aplint32	*from;
	aplcmpx128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLCMPX, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		to->im = 0.0, to->re = *from;

	return(dest);
}

static ARRAY *zget_d(ARRAY *src)	// get APLCMPX from APLDOUB
{
	aplfloat64	*from;
	aplcmpx128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLCMPX, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		to->im = 0.0, to->re = *from;

	return(dest);
}

static ARRAY *zget_D(ARRAY *src)	// get APLCMPX from APLDECF
{
	apldecf128	*from;
	aplcmpx128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLCMPX, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		{
		aplfloat64 d;

		to->im = 0.0;
		to->re = d = decf_to_double(*from);
		if(isinf(d))
			{
			freearray(dest);
			return adomx(src);
			}
		}

	return(dest);
}

static ARRAY *Dget_b(ARRAY *src)	// get APLDECF from APLBOOL
{
	apluns8		*from;
	apldecf128	*to;
	BOUND		bnd;
	apluns8		boolmask;

	from = (apluns8 *)arraystart(src);
	to = (apldecf128 *)mk_pt(APLDECF, src->rank, src->shape);
	bnd = els;

	for(--from, boolmask = 0; bnd--; ++to)
		{
		if(! (boolmask >>= 1))
			{
			boolmask = 0x80;
			++from;
			}
		*to = (boolmask & *from) ? DECF_ONE : DECF_ZERO;
		}
	return(dest);
}

static ARRAY *Dget_s(ARRAY *src)	// get APLDECF from APLSINT
{
	aplint8		*from;
	apldecf128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDECF, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		*to = decf_from_int32(*from);

	return(dest);
}

static ARRAY *Dget_i(ARRAY *src)	// get APLDECF from APLINTG
{
	aplint16	*from;
	apldecf128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDECF, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		*to = decf_from_int32(*from);

	return(dest);
}

static ARRAY *Dget_l(ARRAY *src)	// get APLDECF from APLLONG
{
	aplint32	*from;
	apldecf128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDECF, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		*to = decf_from_int32(*from);

	return(dest);
}

static ARRAY *Dget_d(ARRAY *src)	// get APLDECF from APLDOUB
{
	aplfloat64	*from;
	apldecf128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDECF, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		*to = decf_from_double(*from);

	return(dest);
}

static ARRAY *dget_D(ARRAY *src)	// get APLDOUB from APLDECF
{
	apldecf128	*from;
	aplfloat64	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDOUB, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		{
		aplfloat64 d;

		*to = d = decf_to_double(*from);
		if(isinf(d))
			{
			freearray(dest);
			return adomx(src);
			}
		}

	return(dest);
}

static ARRAY *Dget_B(ARRAY *src)
{
	apldecf128	*from;
	apldecf128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDECF, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		*to = decf_bid_to_dpd(*from);

	return(dest);

}

static ARRAY *dget_B(ARRAY *src)	// get APLDOUB from APLDECF
{
	ARRAY *dpd = Dget_B(src);
	ARRAY *doubarray = dget_D(dpd);	// FIXME if this errors on overflow dpd won't get released

	freearray(dpd);
	return doubarray;
}

static ARRAY *Bget_D(ARRAY *src)
{
	apldecf128	*from;
	apldecf128	*to;
	BOUND		bnd;

	from = arraystart(src);
	to = mk_pt(APLDECF, src->rank, src->shape);
	bnd = els;

	for(; bnd--; ++to, ++from)
		*to = decf_dpd_to_bid(*from);

	return(dest);

}

static ARRAY *BIDget(ARRAY *src, ARRAY *(*dpdfn)(ARRAY *))
{
	void  *dontcare = (*dpdfn)(src);
	ARRAY *dpd = dest;
	ARRAY *bid = Bget_D(dpd);

	freearray(dpd);
	return bid;
}

static ARRAY *Bget_b(ARRAY *src)
{
	return BIDget(src, Dget_b);
}

static ARRAY *Bget_s(ARRAY *src)
{
	return BIDget(src, Dget_s);
}

static ARRAY *Bget_i(ARRAY *src)
{
	return BIDget(src, Dget_i);
}

static ARRAY *Bget_l(ARRAY *src)
{
	return BIDget(src, Dget_l);
}

static ARRAY *Bget_d(ARRAY *src)
{
	return BIDget(src, Dget_d);
}

ARRAY	*(*convfns[NELTYPES][NELTYPES])(ARRAY *) =
{			// <to>get_<from>
// FROM     TO ---->
// |        CHAR   BOOL    SINT    INTG    LONG    DOUB    PNTR    WC8     WC16    WC32	     CMPX    RATS    DPD     BID
/*CHAR	*/ {aget_a,adomx,  adomx,  adomx,  adomx,  adomx,  anonx  ,anonx, anonx,   anonx,    adomx,  adomx,  adomx,  adomx },
/*BOOL	*/ {adomx, aget_a, sget_b, iget_b, lget_b, dget_b, anonx  ,adomx, adomx,   adomx,    zget_b, adomx,  Dget_b, Bget_b},
/*SINT	*/ {adomx, anonx , aget_a, iget_s, lget_s, dget_s, anonx  ,adomx, adomx,   adomx,    zget_s, adomx,  Dget_s, Bget_s},
/*INTG  */ {adomx, anonx , anonx , aget_a, lget_i, dget_i, anonx  ,adomx, adomx,   adomx,    zget_i, adomx,  Dget_i, Bget_i},
/*LONG	*/ {adomx, anonx , anonx , anonx , aget_a, dget_l, anonx  ,adomx, adomx,   adomx,    zget_l, adomx,  Dget_l, Bget_l},
/*DOUB	*/ {adomx, anonx , anonx , anonx , anonx , aget_a, anonx  ,adomx, adomx,   adomx,    zget_d, adomx,  Dget_d, Bget_d},
/*PNTR	*/ {anonx, anonx , anonx , anonx , anonx , anonx , anonx  ,anonx, anonx,   anonx,    adomx,  adomx,  adomx,  adomx },
/*WC8	*/ {adomx, adomx,  adomx,  adomx,  adomx,  adomx,  anonx  ,aget_a,c8getc16,c8getc32, adomx,  adomx,  adomx,  adomx },
/*WC16	*/ {adomx, adomx,  adomx,  adomx,  adomx,  adomx,  anonx  ,adomx, aget_a,  c16getc32,adomx,  adomx,  adomx,  adomx },
/*WC32	*/ {adomx, adomx,  adomx,  adomx,  adomx,  adomx,  anonx  ,adomx, adomx,   aget_a,   adomx,  adomx,  adomx,  adomx },
/*CMPX  */ {adomx, anonx , anonx , anonx , anonx , anonx,  anonx  ,adomx, adomx,   adomx,    aget_a, adomx,  adomx,  adomx },
/*RATS  */ {adomx, adomx , adomx , adomx , adomx , adomx,  adomx  ,adomx, adomx,   adomx,    adomx,  aget_a, adomx,  adomx },
/*DPD   */ {adomx, anonx , anonx , anonx , anonx , dget_D, anonx  ,adomx, adomx,   adomx,    zget_D, adomx,  aget_a, Bget_D},
/*BID   */ {adomx, anonx , anonx , anonx , anonx , dget_B, anonx  ,adomx, adomx,   adomx,    anonx,  adomx,  Dget_B, aget_a}
};
