/******************************************************************************
*  affirm macros
******************************************************************************/

#ifndef AFFIRM_H
#define AFFIRM_H

#if !AP_BUILD

static FORCEINLINE int
_sizechk (unsigned int sz, int sgn, uintmax_t v, int neg)
{
	uintmax_t m;
	if (!sgn && neg) return 0;
	m = ~(uintmax_t)0 >> (sizeof (uintmax_t) * 8 - sz * 8 + sgn);
	if (neg) v = ~v;
	return (v & ~m) == 0;
}

// Care: V is evaluated multiple times
#define SAFE_TO_CAST(T,V)						      \
	(_sizechk((unsigned int)sizeof(T),(T)-1<0,			      \
	(uintmax_t)(V),(V)<0))

#if _DEBUG

extern void _affirm(char *msg);

#define affirm(C)							      \
	((C)?(void)0:_affirm(__FILE__":"				      \
	TOSTRING(__LINE__) ": \"" TOSTRING(C) "\" is false"))

#else // _DEBUG

#define affirm(C)		((void) 0)

#endif // _DEBUG

#else // !AP_BUILD

#define SAFE_TO_CAST(T,V)	1
#define affirm(C)		((void) 0)

#endif // !AP_BUILD

#define CAST(T,V)		(affirm(SAFE_TO_CAST(T,V)),(T)(V))
#define ECAST(T,V,E)		(SAFE_TO_CAST(T,V)?(T)(V):(error(E),0))
#define LCAST(T,V)		ECAST(T,V,ELIMIT)

#endif // AFFIRM_H

//              Copyright (c) 1982 Dyadic Systems Limited
