#include <machine.h>
#include <errno.h>
#include <stdio.h>
#if	UNIX
#	include <unistd.h>
#else	//UNIX
#	include	<io.h>
#endif	//UNIX
#include <stdlib.h>
#include <error.h>
#include <win_io.h>

#define MINOF(a, b)	((a) <= (b) ? (a) : (b))

#ifndef APLTIME_DEFAULT
#define APLTIME_DEFAULT 2
#endif	// APLTIME_DEFAULT 

#ifdef	XAPL
int		jwm_proc_key;
#endif	// XAPL 

#if defined(MAINWIN) || WINE
#	if	GUI
#		undef BELL
#		undef BSD
#	endif//GUI
#endif // MAINWIN 

#ifdef	BELL
#include <fcntl.h>
#if	defined(LINUX)
#include <termios.h>
#else //defined(LINUX)
#include <termio.h>
#endif//defined(LINUX)
#include <sys/ioctl.h>
#endif	// BELL 

#ifdef	BSD
#include <sys/time.h>
#endif	// BSD 

#include <jumps.h>
#include <writloss.h>

#include <unix.h>
#if	AP_BUILD
	extern gchar* W3getenv(const gchar *envname);
	extern void W3Error(gchar *buf,...);
	extern int W3_timed_read(int fd,int wait);
	extern void W3_timed_ansi(int fd,int wait,int *value,int *ansi);
	extern void trans_message_box(gchar *message);
	extern size_t getcmdprofile(const gchar *name,gchar * buf,size_t buflen);
	extern size_t getglobalenv( const gchar *name,gchar * buf,size_t buflen);
#else //AP_BUILD
#	include <w3fsio.h>
#	include <embed.h>
#	include <util.h>
#	include <thread.h>
#endif//AP_BUILD

#include 	<ride.h>

static size_t uninterrupted_read(int fd, UCH *buf, size_t wanted);
extern int indq(void);

#ifdef	BSD

int timed_read(fd, wait)
int fd;
int wait;
{
	/*
		some machines show problems here especially when the
		keyboard is connected remotely via a network.
		In these circumstances it is not possible to set a time
		in which we can guarrentee to have collected all of the
		characters transmitted by a single key stroke.
	Alpha error:
		Users single keystroke is interpretted as multiple key strokes.
	Beta error:
		Users multiple key strokes are interpretted as a single stroke.

		The intention is to err on the side of the beta error on the
		basis that there is a low probability of actually typing
		the character sequence transmitted by a special key. Some
		terminals do however use 'ESC A' which is a common thing
		to type at vi.
	*/
	static struct timeval timeout;
	static UCH time_out;		// delay in 1/10ths sec		
	fd_set input;
	UCH ch;

	FD_ZERO(&input);
	FD_SET(fd,&input);

#ifdef XAPL
	if(isanX(fd))
		{
		int	rslt;
		jwm_proc_key = 0;		// not processing 
		rslt =  X_timed_read(fd, wait);
		jwm_proc_key = 1;		// go'n'process key 
		return rslt;
		}
#endif // XAPL 

	if(!time_out)	// first time through 
		{
		UCH *ep;

		ep = (void *)getenv("APLTIME");
		if(ep)time_out = atoi(ep);
		if(!time_out)time_out = APLTIME_DEFAULT;
		timeout.tv_sec = time_out / 10;
		timeout.tv_usec = (time_out % 10) * 100000;
		}

START: 

#if UNIX && !AP_BUILD
	{
	achar a;
	if (a=next_ride_char())
		{
		count=0;
		return a;
		}
	}
#endif

	if (wait)
		{
#if	AP_BUILD
#else //AP_BUILD
		if(0 == pipelen(fd))
			{
			errno = 0;
			tswitch_at_descalc(fd, TSWITCH_ON);
			if
				(
				strong
			||	(weak && indq())
				)switch(errno)
				{
			/* case EINTR: */ #include	<eintr.h>
				return -3;	// EINTR from the select()
				}
			}
		tswitch_at_descalc(fd, TSWITCH_OFF);
#endif//AP_BUILD
		}

		if (1==uninterrupted_read(fd, &ch, 1))
			{
			if (ch==0) goto START;

			return (int)ch & 0xff;
			}
		else
			{
			switch(errno)
				{
			default:
				return -1;	//EOF
			/* case EINTR: */ #include	<eintr.h>
				break;
				}
			return -3;	// INTERRUPTED 
			}

	return -2;			// TIMEOUT 
}

#endif	// BSD 

#ifdef	BELL

/*
 *	The BELL (System III or System V) version is much trickier.
 *	Firstly, if 'wait' is true then we can just issue a read for
 *	one character (the caller must already have set -ICANON if
 *	stdin is a tty).
 *	Otherwise, ('wait' is false) we have three alternative methods
 *	of implementing a non-blocking read.
 *	(1)	If there is a way of finding how many characters are
 *		queued we can use this to find whether a read would
 *		block or not, and if there are some characters ready
 *		then we just issue a read.
 *		NOTE: this section of code is currently commented out
 *		because it doesn't work on the CADMUS and the ALTOS.
 *		On the CADMUS, using FIONREAD causes characters to be
 *		lost; on the ALTOS, TCRDCHK appears to return a random
 *		number 0..15.
 *	(2)	We set O_NDELAY mode and issue a read.
 *	(3)	If stdin is a tty, then as a last resort we set MIN to
 *		0 and TIME to 0.1 sec, and issue a read.  This method
 *		may be necessary on a slow TTY line, where the read with
 *		O_NDELAY set may report no characters because the next
 *		character of the key sequence has not yet arrived from
 *		the terminal.  This method allows up to 100 ms between
 *		characters before a sequence is considered finished.
 */


#ifndef TCSETOP			// default if not defined above		
#define NBUF	1		// buffer size 1 gives smooth response	
#define TCSETOP TCSETA		// use TCSETA unless it gives trouble	
#endif	// TCSETOP 

int timed_read(int fd, int wait)
{
	static UCH time_out;		// delay in 1/10ths sec		
	static UCH	buff[NBUF], *bfptr;
	static size_t		count = 0;
	int			flagcontext;
	struct termio		ttycontext,
				ttynodelay;
	int fail = -1;	// set to EOF until we are sure we fail on a tty 

#ifdef XAPL
	if(isanX(fd))
		{
		int	rslt;
		jwm_proc_key = 0;		// not processing 
		rslt =  X_timed_read(fd, wait);
		jwm_proc_key = 1;		// go'n'process key 
		return rslt;
		}
#endif // XAPL 

	if(count)
	{
		--count;
		return *bfptr++;
	}

START: 
	count=count;

#if UNIX && !AP_BUILD
	{
	achar a;
	if (a=next_ride_char())
		{
		count=0;
		return a;
		}
	}
#endif

	if(wait)				// willing to wait	
	{
#if	AP_BUILD
#else //AP_BUILD
		if(0 == pipelen(fd))
			{
			errno = 0;
			tswitch_at_descalc(fd, TSWITCH_ON);
			if
				(
				strong
			||	(weak && indq())
				)switch(errno)
				{
			/* case EINTR: */ #include	<eintr.h>
				count = 0;
				goto DONE;	// EINTR from the select()
				}
			}
		tswitch_at_descalc(fd, TSWITCH_OFF);
#endif//AP_BUILD
		count = uninterrupted_read(fd, buff, 1); 
		if (count==1 && buff[0]==0) goto START;
		goto DONE;
	}

	flagcontext = fcntl(fd, F_GETFL, 0);
	fcntl(fd, F_SETFL, flagcontext|O_NDELAY);
	count = uninterrupted_read(fd, buff, NBUF);// read with no delay
if (count==1 && buff[0]==0) goto START;
	fcntl(fd, F_SETFL, flagcontext);
	if(count)				// got some characters	
		goto DONE;

	if(ioctl(fd, TCGETA, &ttycontext) == 0)	// then it's a tty	
	{
		fail = -2;	// set to TIMEOUT 
		if(!time_out)	// first time through 
			{
			UCH *ep;

			ep = (void *)getenv("APLTIME");
			if(ep)time_out = atoi((char*)ep);
			if(!time_out)time_out = APLTIME_DEFAULT;
			}
		ioctl(fd, TCGETA, &ttynodelay);	// copy tty structure	
		ttynodelay.c_lflag &= ~ICANON;	// make sure ICANON off 
		ttynodelay.c_cc[VMIN] = 0;		// minimum characters	
		ttynodelay.c_cc[VTIME] = time_out;	// timeout		
		ioctl(fd, TCSETOP, &ttynodelay); // set nodelay modes	
		count = uninterrupted_read(fd, buff, NBUF);// get some chars
		ioctl(fd, TCSETOP, &ttycontext); // restore normal modes 
	}

DONE:
	bfptr = buff;
	if(count > 0)
		return  (int)(--count, *bfptr++) & 0xff;
	switch(errno)
		{
	default:
		return fail;
	/* case EINTR: */ #include	<eintr.h>
		break;
		}
	return -3;
}

#endif	// BELL 


#if	GUI && defined(WINDOWS3)

int	timed_read(int fd, int wait)
	{
	int c = -2;	// mark it as time out 

	if
		(
		1
#if	AP_BUILD || APL_IME
#else //AP_BUILD || APL_IME
	&&	0 == RuntimeIsConsole()
#endif//AP_BUILD || APL_IME
	&&	is_user_io(fd)
		)
		{
		c = W3_timed_read(fd, wait);
		if(EOF == c) return -1;
		}
	else
		{
#if USES_IME && !AP_BUILD && defined(WINDOWS3)
		c = WindowsFileRead(fd);
#else
			{
			UCH t;
			if(0 == read( fd, &t, 1)) 
				return -1;
			c = t;
			}
#endif
		}
	return c;
	}
#endif//GUI && defined(WINDOWS3)

#if defined(UNIX)
static size_t uninterrupted_read(int fd, UCH *buf, size_t wanted)
{
ssize_t count;

	errno = 0;

#if !AP_BUILD && HAS_RIDE
	if ((count=process_ride(fd,buf,wanted)) <= 0)
#endif
		count = read(fd, buf, wanted);
	return count<0 ? 0 : count;
}
#endif
