//	Copyright (c) 1982 Dyadic Systems Limited	

#include <machine.h>
#include <quadav.h>
#include <stddef.h>
#include <trans.h>
#include <resolve.h>
#include <pockutil.h>

#include <string.h>

static unsigned res(unsigned a, unsigned b);

unsigned resolve(unsigned a, unsigned b)
/*
	Resolve overstrikes
*/
{
 unsigned r;

 if(a==b)return a;
 if(a==AV_SP)return b;
 if(b==AV_SP)return a;
 if(a>b)
	{
	 r = a;
	 a = b;
	 b = r;
	}
 r = res(a, b);
#if	HAS_UNICODE
	// Coding the other way around for the underscored character set would
	// be horrible.
	// Opted for doing the second search instead.
	// Probably better to make this table driven and sort the table on first
	// use.
 return r ? r : res(b, a);
#else //HAS_UNICODE
 return r;	// switch statement is ordered for a classic interpreter
#endif//HAS_UNICODE
}

#define KEYA(a,b) ((a<<16)|b)
#define KEYB(a,b) ((b<<16)|a)


#if	HAS_UNICODE
	#define	UNDERSCORE(b)	((b-AV_A)+0x24b6)
#else //HAS_UNICODE
	#define UNDERSCORE(b)   (b|0x20)
#endif//HAS_UNICODE

static int LIBCALL overstrike_sort_a(const void *v1, const void *v2)
{
const overstrike *o1 = (const overstrike *)v1;
const overstrike *o2 = (const overstrike *)v2;

return (o1->a < o2->a) ? -1
     : (o1->a > o2->a) ?  1
     : 0;
}

static int LIBCALL overstrike_sort_b(const void *v1, const void *v2)
{
const overstrike *o1 = (const overstrike *)v1;
const overstrike *o2 = (const overstrike *)v2;

return (o1->b < o2->b) ? -1
     : (o1->b > o2->b) ?  1
     : 0;
}


overstrike overstrikes[]=
	{
	{AV_QUOTE,AV_POINT,AV_SHRIEK},
        {AV_QUOTE,AV_QUAD,AV_QUADQUOTE},

#if HAS_UNICODE
	{AV_COLON,AV_QUAD, AV_VARIANT},
#endif

#if  !APL_IME
	{AV_UNDER,AV_A, UNDERSCORE(AV_A)},
	{AV_UNDER,AV_B, UNDERSCORE(AV_B)},
	{AV_UNDER,AV_C, UNDERSCORE(AV_C)},
	{AV_UNDER,AV_D, UNDERSCORE(AV_D)},
	{AV_UNDER,AV_E, UNDERSCORE(AV_E)},
	{AV_UNDER,AV_F, UNDERSCORE(AV_F)},
	{AV_UNDER,AV_G, UNDERSCORE(AV_G)},
	{AV_UNDER,AV_H, UNDERSCORE(AV_H)},
	{AV_UNDER,AV_I, UNDERSCORE(AV_I)},
	{AV_UNDER,AV_J, UNDERSCORE(AV_J)},
	{AV_UNDER,AV_K, UNDERSCORE(AV_K)},
	{AV_UNDER,AV_L, UNDERSCORE(AV_L)},
	{AV_UNDER,AV_M, UNDERSCORE(AV_M)},
	{AV_UNDER,AV_N, UNDERSCORE(AV_N)},
	{AV_UNDER,AV_O, UNDERSCORE(AV_O)},
	{AV_UNDER,AV_P, UNDERSCORE(AV_P)},
	{AV_UNDER,AV_Q, UNDERSCORE(AV_Q)},
	{AV_UNDER,AV_R, UNDERSCORE(AV_R)},
	{AV_UNDER,AV_S, UNDERSCORE(AV_S)},
	{AV_UNDER,AV_T, UNDERSCORE(AV_T)},
	{AV_UNDER,AV_U, UNDERSCORE(AV_U)},
	{AV_UNDER,AV_V, UNDERSCORE(AV_V)},
	{AV_UNDER,AV_W, UNDERSCORE(AV_W)},
	{AV_UNDER,AV_X, UNDERSCORE(AV_X)},
	{AV_UNDER,AV_Y, UNDERSCORE(AV_Y)},
	{AV_UNDER,AV_Z, UNDERSCORE(AV_Z)},
#endif // APL_IME

	{AV_UNDER,AV_DELTA, AV_DELTAU},
	{AV_UNDER,AV_EQ, AV_MATCH},
	{AV_UNDER,AV_NE, AV_NOTMATCH},
	{AV_UNDER,AV_EPSILON, AV_EPSUBAR},

	{AV_ZERO,AV_TILDE,AV_ZILDE},
	{AV_L   ,AV_MINUS,AV_POUND} ,
	{AV_S   ,AV_STILE,AV_DOLLAR},

	{AV_DELTA,AV_STILE,AV_UPGRADE},
	{AV_DIRESIS,AV_TILDE, AV_COMM},
	{AV_DIRESIS,AV_STAR,  AV_DIRESISSTAR},

	{AV_LBRACKET,AV_JOT	, AV_LBRACE},
	{AV_LBRACKET,AV_RBRACKET, AV_SQUISH},

	{AV_SLASH,AV_MINUS	, AV_BSLASH},

	{AV_SLOPE,AV_MINUS	, AV_BSLOPE},
	{AV_SLOPE,AV_CIRCLE	, AV_TRANSPOSE},

	{AV_LT,AV_GT	, AV_DIAMOND},

	{AV_OR,AV_AND	, AV_DIAMOND},
	{AV_OR,AV_TILDE	, AV_NOR},

	{AV_AND,AV_TILDE	, AV_NAND},

	{AV_MINUS,AV_CIRCLE	, AV_BCIRCLE},
#if !APL_IME
	{AV_MINUS,AV_STILE	, AV_BSTILE},
#endif
	{AV_MINUS,AV_COMMA	, AV_BCOMMA},
	{AV_MINUS,AV_LPAREN	, AV_RTACK},
	{AV_MINUS,AV_RPAREN	, AV_LTACK},
	{AV_MINUS,AV_E	, AV_EURO},

	{AV_DIVIDE,AV_QUAD	, AV_DOMINO},

	{AV_TILDE,AV_DEL	, AV_DELTILDE},

	{AV_CIRCLE,AV_STAR	, AV_LOG},
	{AV_CIRCLE,AV_STILE	, AV_CIRCLESTILE},

	{AV_DEL,AV_STILE	, AV_DGRADE},

	{AV_JOT,AV_CAP	, AV_LAMP},
	{AV_JOT,AV_BASE	, AV_HYDRANT},
	{AV_JOT,AV_TOP	, AV_THORN},
	{AV_JOT,AV_RBRACKET	, AV_RBRACE},

	{AV_LSHOE,AV_STILE	, AV_CENT},
	{AV_BASE,AV_TOP	, AV_IBEAM},
};

#define SOVERSTRIKE  (sizeof(overstrikes[0]))
#define NOVERSTRIKES (sizeof(overstrikes)/SOVERSTRIKE)

overstrike overstrikes_a[NOVERSTRIKES];
overstrike overstrikes_b[NOVERSTRIKES];

static unsigned res(unsigned a, unsigned b)
{
static int init;
overstrike target,*found;

if (init==0)
	{
	size_t i;

	for (i=0;i<NOVERSTRIKES;i++)
		{
		overstrikes[i].a=KEYA(overstrikes[i].first,overstrikes[i].second);
		overstrikes[i].b=KEYB(overstrikes[i].first,overstrikes[i].second);
		}

	memcpy(overstrikes_a,overstrikes,sizeof(overstrikes_a));
	memcpy(overstrikes_b,overstrikes,sizeof(overstrikes_b));

	qsort(overstrikes_a,NOVERSTRIKES,SOVERSTRIKE,overstrike_sort_a);
	qsort(overstrikes_b,NOVERSTRIKES,SOVERSTRIKE,overstrike_sort_b);

	init=1;
	}

target.a=KEYA(a,b);

if (found=bsearch(&target,overstrikes_a,NOVERSTRIKES,SOVERSTRIKE,overstrike_sort_a))
	return found->resolved;

return 0;
}

size_t overstrike_candidates(overstrike *buf, size_t blen, unsigned c)
{
size_t i,n=0;
overstrike *bp=buf;

for (i=0;i<NOVERSTRIKES;i++)
	{
	if (overstrikes[i].first==c || overstrikes[i].second==c)
		{
		if (n==blen) break;
		bp[n++]=overstrikes[i];
		}
	}
return n;
}



#if !APL_IME

unsigned convmode;	// exit mode from last convtoav call 

unsigned convtoav
	(
	achar *to		// destination string	
,	void *from	// source string	
,	uns eltype
,	size_t max	// max characters to transfer 
	)
{
/*
	copies from source to destination translating to av
	returns the length of the resulting string
	if max is less than the value returned the the copy
	will be incomplete
	note: tab characters can make the output bigger than the input
*/
register unsigned count=0,	// count of output characters 
		  bscount=0;	// count of unused backspaces 
register achar c, new;
register unsigned aplmode=0;	// initial mode of input 0 implies ascii
				//			 1 implies apl	
size_t ifp=0;

while(c=nthachar_et(from,eltype,ifp++))
	{
	 if(32>c)switch(c)	// deal with control characters	
		{
		 case '\b':		// back space	
			 if(bscount<count)
				{
				 ++bscount;
				 to--;
				}
			 continue;
		 case '\t':		// tab		
				{
				 unsigned newcount;

				 newcount=(count-bscount+8)& ~7;
				 if(newcount>count)
					{
					 to+=bscount;
					 bscount=0;
					 while(count<newcount)
						{
						 if(count++ < max)*to = AV_SP;
						 ++to;
						}
					}
				 else 
					{
					 newcount=count-newcount;
					 to+=bscount-newcount;
					 bscount=newcount;
					}
				}
			 continue;
		 case '\n':		// newline	
		 case '\r':		// cr		
			 new=AV_NEWLINE;
			 break;
#if	HAS_UNICODE
#else //HAS_UNICODE
		 case '\016':
			 aplmode=1;
			 continue;
		 case '\017':
			 aplmode=0;
			 continue;
#endif//HAS_UNICODE
		 default:new=AV_SQUISH;	// unrecognised control char 
			 break;
		}
	 else new=asctoav(c,aplmode);
	 if(!bscount)
		{
		 if(count<max)*to = new;
		 ++count;
		}
	 else
		{
		 if((count-bscount) < max)
			{
			 if(0==(c=resolve(*to,new)))c=new;
			 *to = c;
			}
		 bscount--;
		}
	 ++to;
	}
convmode=aplmode;	// set global exit mode 
return count;
}

size_t convlen(void *from,uns eltype)
{
/*
	find length of converted string without copying anything
*/
return convtoav((achar *)0, from, eltype,(size_t)0);
}

#endif // !APL_IME

//	Copyright (c) 1982 Dyadic Systems Limited	

