
/*
 * Implement unix type calls
 */
#include <machine.h>
#if	UNIX
#	include <unistd.h>
#else //UNIX
#	include <io.h>
#endif//UNIX

#if	WINCE
#	include <wince.h>
#elif	defined(WINDOWS3)
#	include <windows.h>
#endif // WINCE

#include <gfns.h>

#if !APL_IME

static UCH forced_user_io[3] = {0, 0, 0};


#if defined(DOSWIN) 
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <dos.h>
#include <dosfns.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <runtime.h>

static int fileclass(int fd);

/*
 * dummy out calls which we hope not to use at the moment
 */

int alarm(void){return 0;}
intptr_t LIBCALL execv(const char *path, char *const *argv){nonce(); return 0;}
intptr_t LIBCALL execl(const char *path, const char *arg,...){nonce();return 0;}
void fork(void){nonce();}

#if (!defined(MSC))
void locking(void){nonce();}
void mktemp(void){nonce();}
#endif

int pause(void){nonce(); return -1;}
void sync(void){}
int wait(int *stat_loc){nonce(); return -1;}
void slave(int *fd, UCH *path, UCH *cmd, UCH **params)
	{
	nonce();
	}
void kill(int pid, int sig){nonce();}
int REMOTEFILE(int st_dev){return 0;}

/* implement easy functions */


char *ttyname(int fd)
{
switch(fileclass(fd))
	{
	case 1: case 2: case 3:
		if(forced_user_io[fd] || isatty(fd))
			return "/dev/console";
	default:
		return 0;
	}
}

/* forget about cpu times */

void times(void){}

/* uname is MS_DOS */

char *uname() { return "MS_DOS"; }

int32_t getuid()
{
	static unsigned first = 1;
	static int32_t uid;

	if(first)
		{
		first = 0;
		uid = atol((void *)getenv("APLNID"));
		}
	return uid;
}

#if !defined(WINCE)
void binary_file(int fd)	/* make file look unixy */
{
	setmode(fd, O_BINARY);
}

void text_file(int fd)	/* make file look dosy */
{
	setmode(fd, O_TEXT);
}
#endif


void binary_stream(FILE *stream)
{
#if defined(WINCE)
	setmode(stream,O_BINARY);
#else
	binary_file(fileno(stream));
#endif
}


void text_stream(FILE *stream)
{
#if defined(WINCE)
	setmode(stream,O_TEXT);
#else
	text_file(fileno(stream));
#endif
}

#if WRITLOSS
int non_writloss_fsync_fn(int fd)
#else // WRITLOSS
int fsync(int fd)
#endif //WRITLOSS
	{
	dos_commit_file(fd);
	return 0;
	}


#endif//DOSWIN

#if	defined(UNIX)
#	include <sys/stat.h>
#	include <errno.h>
#endif//defined(UNIX)

static int fileclass(int fd)
	{
	/*	These comments are a guide only - don't trust them
	 * returns :
	 *	-1<-> invalid handle etc
	 *	0 <-> disk
	 *	1 <-> keyboard
	 *	2 <-> screen
	 *	3 <-> either keyboard or screen ( two way )
	 *	4 <-> NUL
	 *	5 <-> CLOCK
	 *	15<-> some other non disk (Serial, parallel)
	 */


#if	defined(UNIX)
	{

	struct stat fd_stat;

	if(0 == fstat(fd, &fd_stat))
		{
		unsigned m = fd_stat.st_mode;

		// POSIX standard macros
		if(S_ISREG(m)) return 0;
		if(S_ISCHR(m)) return 3;
		return 15;
		}
	else return -1;
	}
#elif	defined(__NT__)
	{
	int dx = get_device_info(fd);

	switch (dx)
		{
	case 0: return 3;	// FILE_TYPE_UNKNOWN
	case 1: return 0;	// FILE_TYPE_DISK
	case 2:	return 3;	// FILE_TYPE_CHAR
	case 3:	return 15;	// FILE_TYPE_PIPE
		}
	return -1;
	}
#else
	{
	unsigned dx = get_device_info(fd);

	switch(dx & 0x80)
		{
	case 0:	return 0;
		}
	switch (dx & 0x7)
		{
	case 1: return 1;
	case 2: return 2;
	case 3: return 3;
	case 4: return 4;
	case 8: return 5;
		}
	return 15;
	}
#endif
	}


void force_user_io(int fd)
{
	forced_user_io[fd] = 1;
}


int is_user_io(int fd)
	{
#if	defined(UNIX)

	return isatty(fd);

#else //defined(UNIX)

	switch(fd)
		{
#if	GUI && defined(WINDOWS3)
	case 0: case 1: case 2:
		return
			forced_user_io[fd]
		||	isatty(fd)
		||	3 == fileclass(fd)
		;
#endif//GUI && defined(WINDOWS3)
	default:
		return isatty(fd);
		}

#endif//defined(UNIX)
	}

int isadisk(int fd)
	{
#if	WINCE
	return 1;
#else //WINCE
	return !fileclass(fd);
#endif//WINCE
	}

#endif //APL_IME

#define DEFINE_UNIQUE_FID
#include <unique_fid.h>

/*	This is an attempt to solve a particular problem. Given two file
 *	descriptors, do they refer to the same file?
 *	In addition I want to support a delayed comparison. Take the information
 *	for one handle at one time. Compare it with the information for another
 *	handle (possibly the same one) later.
 *	UNIX:
 *	On a UNIX system this is straight forward: call fstat() on each and
 *	if the st_dev and st_ino fields match it is the same file.
 *	WINDOWS:
 *	On windows systems this is much more complicated. The complications
 *	arise from several areas:
 *		Pipes and files use different calls.
 *		Win 9x, WinNT, WinCE have different support for those calls.
 *		GetFileInformationByHandle() leaves a lot undocumented. When the
 *		file is on a network I may not get all of the information I
 *		would like.
 *		GetFileInformationByHandle() requires a handle. CreateFile()
 *		fails to open directories on win 9x. Thus I can't get the 
 *		information for a directory on 9x.
 *	I can get very little information about a pipe. All of the information
 *	is likely to be the same anyway. Thus I have to assume that when
 *	comparing file descriptors one pipe is the same pipe as another. This
 *	will solve the "2>&1 |" syntax in ksh or cmd.exe. It will not solve
 *	issues with 1>named_pipe_1 2>named_pipe_2. Which is a pity but probably
 *	doesn't affect anybody.
 */

#if	WINCE

	static unsigned wince_count = 0;

#elif	defined(UNIX)
#elif	defined(__NT__)

typedef BOOL (__stdcall *GETFILEINFORMATIONBYHANDLEP)
	(
	HANDLE hFile
,	LPBY_HANDLE_FILE_INFORMATION lpFileInformation
	);

static GETFILEINFORMATIONBYHANDLEP GetFileInformationByHandleP;

static BOOL __stdcall GetFileInformationByHandle95
	(
	HANDLE hFile
,	LPBY_HANDLE_FILE_INFORMATION lpFileInformation
	)
	{
	DWORD low;
	DWORD high;

	return GetFileTime
		(
		hFile
	,	0
	,	0
	,	&lpFileInformation->ftLastWriteTime
		)
	&&	0 != memcpy		// dirty 2 DWORDS make a FILETIME
		(
		&lpFileInformation->nFileIndexHigh // and nFileIndexLow
	,	&lpFileInformation->ftLastWriteTime
	,	sizeof(FILETIME)
		)
	&&	INVALID_FILE_SIZE != (low = GetFileSize(hFile, &high))
	&&	(lpFileInformation->nFileSizeHigh = high, 1)
	&&	(lpFileInformation->nFileSizeLow  = low,  1)
	;
	}

static void handle_to_unique_fid
	(
	HANDLE h
,	UNIQUE_FID *ufidp
,	FID_TYPE (*fillfn)
		(
		UNIQUE_FID *ufidp
	,	BY_HANDLE_FILE_INFORMATION *infp
		)
	)
	{
	BY_HANDLE_FILE_INFORMATION inf;
	static unsigned first = 1;
	FID_TYPE type = FID_INVALID;

	if(first)
		{
		HMODULE hm = LoadLibraryA("kernel32");

		GetFileInformationByHandleP = (GETFILEINFORMATIONBYHANDLEP)GetProcAddress
			(
			hm
		,	"GetFileInformationByHandle"
			);

		FreeLibrary(hm);
		if(0 == GetFileInformationByHandleP)
			GetFileInformationByHandleP = GetFileInformationByHandle95;
		first = 0;
		}

	switch(GetFileType(h))
		{
	case FILE_TYPE_UNKNOWN:
		if(0 == (ufidp->info.invalid_info.last_error = GetLastError()))
			type = FID_UNKNOWN;
		break;
	case FILE_TYPE_REMOTE:
		type = FID_REMOTE;
		break;
	case FILE_TYPE_PIPE:
		type = FID_PIPE;
		break;
	case FILE_TYPE_DISK:
		if(0 == (*GetFileInformationByHandleP)(h, &inf))
			{
			ufidp->info.invalid_info.last_error = GetLastError();
			break;
			}
		type = (*fillfn)(ufidp, &inf);			// succeeded
		break;
	case FILE_TYPE_CHAR:
		type = FID_CHAR;
		break;
		}
	ufidp->fid_type = type;
	}

static FID_TYPE fill_fd
	(
	UNIQUE_FID *ufidp 
,	BY_HANDLE_FILE_INFORMATION *infp
	)
	{			
#	define	mc(a)	ufidp->info.file_fd_info.a = infp->a

	mc(dwVolumeSerialNumber);
	mc(nFileIndexHigh);
	mc(nFileIndexLow);
	mc(nFileSizeHigh);
	mc(nFileSizeLow);

#	undef	mc
	return FID_FILE_FD;
	}

static FID_TYPE fill_name
	(
	UNIQUE_FID *ufidp 
,	BY_HANDLE_FILE_INFORMATION *infp
	)
	{			
#	define	mc(a)	ufidp->info.file_name_info.a = infp->a

	mc(dwVolumeSerialNumber);
	mc(ftLastWriteTime);
	mc(nFileSizeHigh);
	mc(nFileSizeLow);

#	undef	mc
	return FID_FILE_NAME;
	}

#endif//WINCE || defined(__NT__)

UNIQUE_FID fd_to_unique_fid(int fd)
	{
	UNIQUE_FID ufid = { 0 };	// zap any allignment bytes and unused

#	if	UNIX

	struct stat stat;

	if (-1 == fstat(fd, &stat))
		{
		ufid.fid_type = FID_INVALID;
		ufid.info.invalid_info.last_error = errno;
		}
	else
		{
		ufid.fid_type = FID_FILE;
		ufid.info.file_info.st_dev = stat.st_dev;
		ufid.info.file_info.st_ino = stat.st_ino;
		}

#	elif	WINCE

	switch(fd)
		{
	case -1:
		ufid.fid_type = FID_INVALID;
		break;
	case 0:
		ufid.fid_type = FID_STDIN;
		break;
	case 1: case 2:
		ufid.fid_type = FID_STDOUT;
		break;
	default:
		ufid.fid_type = FID_FILE;
		ufid.info.file_info.count = ++wince_count;
		break;
		}

#	elif	defined(__NT__)

	handle_to_unique_fid((HANDLE)_get_osfhandle(fd), &ufid, fill_fd);

#	else
#		error	not UNIX or NT
#	endif //UNIX || defined(__NT__)

	return ufid;
	}

UNIQUE_FID filename_to_unique_fid(gchar *name)
	{
	UNIQUE_FID ufid = { 0 };	// zap any allignment bytes

#	if	UNIX

	struct stat s_stat;

	if (-1 == gstat(name, &s_stat))
		{
		ufid.fid_type = FID_INVALID;
		ufid.info.invalid_info.last_error = errno;
		}
	else
		{
		ufid.fid_type = FID_FILE;
		ufid.info.file_info.st_dev = s_stat.st_dev;
		ufid.info.file_info.st_ino = s_stat.st_ino;
		}

#	elif	WINCE

	ufid.fid_type = FID_FILE;
	ufid.info.file_info.count = ++wince_count;

#	elif	defined(__NT__)

	HANDLE h = CreateFile
		(
		name
	,	0
	,	FILE_SHARE_DELETE
	|	FILE_SHARE_READ
	|	FILE_SHARE_WRITE
	,	0
	,	OPEN_EXISTING
	,	FILE_ATTRIBUTE_NORMAL
	,	0
		);

	if(INVALID_HANDLE_VALUE == h)
		{
		ufid.fid_type = FID_INVALID;
		ufid.info.invalid_info.last_error = GetLastError();
		}
	else
		{
		handle_to_unique_fid(h, &ufid, fill_name);
		}

#	else
#		error	not UNIX or NT
#	endif //UNIX || defined(__NT__)

	return ufid;
	}

unsigned match_unique_fid(UNIQUE_FID this, UNIQUE_FID that)
	{
	return
		this.fid_type == that.fid_type
	&&	FID_INVALID != this.fid_type
	&&	0 == memcmp
			(
			&this.info
		,	&that.info
		,	sizeof(this.info)
			)
	;
	}

unsigned valid_fd(int fd)
	{
#if	UNIX
	{
	struct stat stat;

	return 0 == fstat( fd, &stat);
	}

#elif	WINCE
	{
	switch(fd)
		{
	case 0: case 1: case 2:
		return 1;
	default:
		return INVALID_HANDLE_VALUE != (HANDLE)fd;
		}
	}
	
#elif	defined(__NT__)

	{
	HANDLE hdl = (HANDLE)_get_osfhandle(fd);

	return hdl != INVALID_HANDLE_VALUE && hdl != (HANDLE)-2;
	}

#else

#	error	not UNIX or NT

#endif//UNIX || defined(__NT__)
	}
